import 'dart:convert';
import 'dart:io';

import 'package:meta/meta.dart';
import 'package:puma/src/package_information.dart';
import 'package:yaml/yaml.dart' as yaml;
import 'package:http/http.dart' as http;
import 'package:pub_semver/pub_semver.dart' as semver;

class PubspecManager {
  final String hostUrl;
  final bool runPubGet;
  File pubspec;
  Map jsonPubspec;

  PubspecManager({@required this.hostUrl, @required this.runPubGet});

  Future<void> init() async {
    pubspec = File('pubspec.yaml');
    final pubspecContent = await pubspec.readAsString();
    final yamlRoot = yaml.loadYaml(pubspecContent) as yaml.YamlMap;
    jsonPubspec = json.decode(json.encode(yamlRoot)) as Map;
  }

  Future<void> install(String package, {bool dev = false}) async {
    final key = !dev ? 'dependencies' : 'dev_dependencies';
    final packageInformations = await _getPackageInformations(package);
    final packageInfos = _generatePackageInformations(packageInformations);
    if (jsonPubspec.containsKey(key)) {
      final dependence = jsonPubspec[key] as Map;
      dependence.addAll(packageInfos);
    } else {
      jsonPubspec.addAll({key: packageInfos});
    }
    await _writeContent(jsonPubspec);
    await pubGet();
  }

  Future<void> delete(String packageName) async {
    var pm = packageName;
    if (packageName.contains(':')) {
      pm = packageName.split(':').first;
    }
    if (_deletePackage('dependencies', pm) ||
        _deletePackage('dev_dependencies', pm)) {
      await _writeContent(jsonPubspec);
      await pubGet();
    } else {
      print('Package $packageName not found');
    }
  }

  Future<void> nextBreakingVersion() async {
    final version = semver.Version.parse(jsonPubspec['version']);
    await _updateVersion(version.nextBreaking);
  }

  Future<void> nextMajorVersion() async {
    final version = semver.Version.parse(jsonPubspec['version']);
    await _updateVersion(version.nextMajor);
  }

  Future<void> nextMinorVersion() async {
    final version = semver.Version.parse(jsonPubspec['version']);
    await _updateVersion(version.nextMinor);
  }

  Future<void> nextPatchVersion() async {
    final version = semver.Version.parse(jsonPubspec['version']);
    await _updateVersion(version.nextPatch);
  }

  Future<void> _updateVersion(semver.Version nextVersion) async {
    jsonPubspec['version'] = nextVersion.toString();
    await _writeContent(jsonPubspec);
    await pubGet();
  }

  Future<void> _writeContent(Map json) async {
    await pubspec.writeAsStringSync(
      '''# GENERATED CODE - DO NOT MODIFY BY HAND

# Pubspec managed entirely with puma
''',
      mode: FileMode.writeOnly,
    );
    await pubspec.writeAsString(
      _mapToYaml(jsonPubspec),
      mode: FileMode.append,
    );
  }

  Future<void> pubGet() async {
    if (!runPubGet) {
      return;
    }
    print('Running pub get for you...');
    final process = await Process.start('pub', ['get']);
    process.stdout.transform(utf8.decoder).listen(stdout.write);
  }

  bool _deletePackage(String dependencesKey, String packageName) {
    final dependences = jsonPubspec[dependencesKey] as Map;
    if (dependences?.containsKey(packageName) == true) {
      dependences?.removeWhere((pn, _) => packageName == pn);
      return true;
    }
    return false;
  }

  Map<String, dynamic> _generatePackageInformations(
      PackageInformation packageInformations) {
    if (packageInformations is HostedPackage) {
      return {
        packageInformations.name: packageInformations.version,
      };
    } else if (packageInformations is PathPackage) {
      return {
        packageInformations.name: {
          'path': packageInformations.path,
        },
      };
    }
    return {};
  }

  Future<PackageInformation> _getPackageInformations(
      String installInformation) async {
    if (installInformation.contains(':')) {
      final colonIndex = installInformation.indexOf(':');
      final packageName = installInformation.substring(0, colonIndex);
      final v = installInformation.substring(colonIndex + 1);
      if (v.startsWith('git')) {
        //  TODO(kevin): load git resource
        throw 'git is not currently supported';
      } else if (v.isNotEmpty && v != 'latest') {
        final _v = semver.VersionConstraint.parse(v);
        if (await _checkVersionExist(packageName, _v)) {
          return HostedPackage(
            packageName,
            _v.toString(),
          );
        } else {
          //  TODO(kevin): add better error handling
          throw 'Version ${_v.toString()} of package $packageName doesn\'t exist';
        }
      } else {
        return HostedPackage(
          packageName,
          await _getLatestStableVersion(packageName),
        );
      }
    } else {
      final normalizePath = installInformation.endsWith(Platform.pathSeparator)
          ? installInformation
          : '$installInformation${Platform.pathSeparator}';
      final pubspecPath = '${normalizePath}pubspec.yaml';
      final otherPubspec = File(pubspecPath);
      final pubspecExist = await otherPubspec.exists();
      if (pubspecExist) {
        final otherYaml =
            yaml.loadYaml(await otherPubspec.readAsString()) as yaml.YamlMap;
        return PathPackage(
          otherYaml['name'],
          normalizePath,
        );
      }
      return HostedPackage(
        installInformation,
        await _getLatestStableVersion(installInformation),
      );
    }
  }

  Future<String> _getLatestStableVersion(String packageName) async {
    final response = await http.get('$hostUrl/api/documentation/$packageName');
    final packageInfos = json.decode(response.body);
    return '^${packageInfos['latestStableVersion']}';
  }

  Future<bool> _checkVersionExist(
      String packageName, semver.VersionConstraint version) async {
    final response = await http.get('$hostUrl/api/documentation/$packageName');
    final packageInfos = json.decode(response.body);
    return (packageInfos['versions'] as List).cast<Map>().firstWhere(
            (d) => version.allows(semver.Version.parse(d['version'])),
            orElse: () => null) !=
        null;
  }

  String _mapToYaml(Map map, {int deep = 0}) {
    final keys = map.keys.toList()..sort();
    final sb = StringBuffer();
    for (final key in keys) {
      final value = map[key];
      if (value is Map) {
        sb.writeln('${' ' * deep}$key:');
        sb.write(_mapToYaml(map[key], deep: deep + 1));
      } else if (value is List) {
        final v = value.toList()..sort();
        sb.writeln('${' ' * deep}$key:');
        sb.write(v.map((String) => '${' ' * (deep + 1)}- $value').join('\n'));
      } else {
        sb.writeln('${' ' * deep}$key: \'$value\'');
      }
    }
    return sb.toString();
  }
}
